"""Custom Sphinx directives."""

import os
import shutil
from collections import namedtuple
from pathlib import Path, PurePosixPath

from docutils import nodes, statemachine
from docutils.parsers.rst import Directive, directives
from docutils.parsers.rst.directives import images
from sphinx.errors import ExtensionError
from sphinx.util.logging import getLogger

from .backreferences import (
    THUMBNAIL_PARENT_DIV,
    THUMBNAIL_PARENT_DIV_CLOSE,
    _thumbnail_div,
)
from .gen_rst import extract_intro_and_title
from .py_source_parser import split_code_and_text_blocks
from .utils import _read_json

logger = getLogger("sphinx-gallery")


class MiniGallery(Directive):
    """Custom directive to insert a mini-gallery.

    The required argument is one or more of the following:

    * fully qualified names of objects
    * pathlike strings to example Python files
    * glob-style pathlike strings to example Python files

    The string list of arguments is separated by spaces.

    The mini-gallery will be the subset of gallery
    examples that make use of that object from that specific namespace

    Options:

    * `add-heading` adds a heading to the mini-gallery.  If an argument is
      provided, it uses that text for the heading.  Otherwise, it uses
      default text.
    * `heading-level` specifies the heading level of the heading as a single
      character.  If omitted, the default heading level is `'^'`.
    """

    required_arguments = 0
    has_content = True
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {
        "add-heading": directives.unchanged,
        "heading-level": directives.single_char_or_unicode,
    }

    def _get_target_dir(self, config, src_dir, path, obj):
        """Get thumbnail target directory, errors when not in example dir."""
        examples_dirs = config.sphinx_gallery_conf["examples_dirs"]
        if not isinstance(examples_dirs, list):
            examples_dirs = [examples_dirs]
        gallery_dirs = config.sphinx_gallery_conf["gallery_dirs"]
        if not isinstance(gallery_dirs, list):
            gallery_dirs = [gallery_dirs]

        gal_matches = []
        for e, g in zip(examples_dirs, gallery_dirs):
            e = Path(src_dir, e).resolve(strict=True)
            g = Path(src_dir, g).resolve(strict=True)
            try:
                gal_matches.append((path.relative_to(e), g))
            except ValueError:
                continue

        # `path` inside one `examples_dirs`
        if (n_match := len(gal_matches)) == 1:
            ex_parents, target_dir = (
                gal_matches[0][0].parents,
                gal_matches[0][1],
            )
        # `path` inside several `examples_dirs`, take the example dir with
        # longest match (shortest parents after relative)
        elif n_match > 1:
            ex_parents, target_dir = min(
                [(match[0].parents, match[1]) for match in gal_matches],
                key=lambda x: len(x[0]),
            )
        # `path` is not inside a `examples_dirs`
        else:
            raise ExtensionError(
                f"minigallery directive error: path input '{obj}' found file:"
                f" '{path}' but this does not live inside any examples_dirs: "
                f"{examples_dirs}"
            )

        # Add subgallery path, if present
        subdir = ""
        if (ex_p := ex_parents[0]) != Path("."):
            subdir = ex_p
        target_dir = target_dir / subdir
        return target_dir

    def run(self):
        """Generate mini-gallery from backreference and example files."""
        from .gen_rst import _get_callables

        if not (self.arguments or self.content):
            raise ExtensionError("No arguments passed to 'minigallery'")

        # Respect the same disabling options as the `raw` directive
        if (
            not self.state.document.settings.raw_enabled
            or not self.state.document.settings.file_insertion_enabled
        ):
            raise self.warning(f'"{self.name}" directive disabled.')

        # Retrieve the backreferences directory
        config = self.state.document.settings.env.config
        backreferences_dir = config.sphinx_gallery_conf["backreferences_dir"]
        if backreferences_dir is None:
            logger.warning(
                "'backreferences_dir' config is None, minigallery "
                "directive will resolve all inputs as file paths or globs."
            )

        # Retrieve source directory
        src_dir = config.sphinx_gallery_conf["src_dir"]

        # Parse the argument into the individual args
        arg_list = []

        if self.arguments:
            arg_list.extend([c.strip() for c in self.arguments[0].split()])
        if self.content:
            arg_list.extend([c.strip() for c in self.content])

        lines = []

        # Add a heading if requested
        if "add-heading" in self.options:
            heading = self.options["add-heading"]
            if heading == "":
                if len(arg_list) == 1:
                    heading = f"Examples using ``{arg_list[0]}``"
                else:
                    heading = "Examples using one of multiple objects"
            lines.append(heading)
            heading_level = self.options.get("heading-level", "^")
            lines.append(heading_level * len(heading))

        ExampleInfo = namedtuple("ExampleInfo", ["target_dir", "intro", "title", "arg"])
        backreferences_all = None
        if backreferences_dir:
            backreferences_all = _read_json(
                Path(src_dir, backreferences_dir, "backreferences_all.json")
            )

        def has_backrefs(arg):
            if backreferences_all is None:
                return False
            examples = backreferences_all.get(arg, None)
            if examples:
                return examples
            return False

        # Full paths to example files are dict keys, preventing duplicates
        file_paths = {}
        for arg in arg_list:
            # Backreference arg input
            if examples := has_backrefs(arg):
                # Note `ExampleInfo.arg` not required for backreference paths
                for path in examples:
                    file_paths[Path(path[1], path[0]).resolve()] = ExampleInfo(
                        target_dir=path[2], intro=path[3], title=path[4], arg=None
                    )
            # Glob path arg input
            elif paths := Path(src_dir).glob(arg):
                # Glob paths require extra parsing to get the intro and title
                # so we don't want to override a duplicate backreference arg input
                for path in paths:
                    path_resolved = path.resolve()
                    if path_resolved in file_paths:
                        continue
                    else:
                        file_paths[path_resolved] = ExampleInfo(None, None, None, arg)

        if len(file_paths) == 0:
            return []

        lines.append(THUMBNAIL_PARENT_DIV)

        # sort on the file path
        if config.sphinx_gallery_conf["minigallery_sort_order"] is None:
            sortkey = None
        else:
            (sortkey,) = _get_callables(
                config.sphinx_gallery_conf, "minigallery_sort_order"
            )
        for path, path_info in sorted(
            file_paths.items(),
            # `x[0]` to sort on key only
            key=((lambda x: sortkey(str(x[0]))) if sortkey else None),
        ):
            if path_info.intro is not None:
                thumbnail = _thumbnail_div(
                    path_info.target_dir,
                    src_dir,
                    path.name,
                    path_info.intro,
                    path_info.title,
                )
            else:
                target_dir = self._get_target_dir(config, src_dir, path, path_info.arg)
                # Get thumbnail
                # TODO: ideally we would not need to parse file (again) here
                _, script_blocks = split_code_and_text_blocks(
                    str(path), return_node=False
                )
                intro, title = extract_intro_and_title(
                    str(path), script_blocks[0].content
                )

                thumbnail = _thumbnail_div(target_dir, src_dir, path.name, intro, title)
            lines.append(thumbnail)

        lines.append(THUMBNAIL_PARENT_DIV_CLOSE)
        text = "\n".join(lines)
        include_lines = statemachine.string2lines(text, convert_whitespace=True)
        self.state_machine.insert_input(include_lines, str(path))

        return []


"""
Image sg for responsive images
"""


class imgsgnode(nodes.General, nodes.Element):
    """Sphinx Gallery image node class."""

    pass


class ImageSg(images.Image):
    """Implements a directive to allow an optional hidpi image.

    Meant to be used with the `image_srcset` configuration option.

    e.g.::

        .. image-sg:: /plot_types/basic/images/sphx_glr_bar_001.png
            :alt: bar
            :srcset: /plot_types/basic/images/sphx_glr_bar_001.png,
                     /plot_types/basic/images/sphx_glr_bar_001_2_00x.png 2.00x
            :class: sphx-glr-single-img

    The resulting html is::

        <img src="sphx_glr_bar_001_hidpi.png"
            srcset="_images/sphx_glr_bar_001.png,
                    _images/sphx_glr_bar_001_2_00x.png 2x",
            alt="bar"
            class="sphx-glr-single-img" />
    """

    has_content = False
    required_arguments = 1
    optional_arguments = 3
    final_argument_whitespace = False
    option_spec = {
        "srcset": directives.unchanged,
        "class": directives.class_option,
        "alt": directives.unchanged,
    }

    def run(self):
        """Update node contents."""
        image_node = imgsgnode()

        imagenm = self.arguments[0]
        image_node["alt"] = self.options.get("alt", "")
        image_node["class"] = self.options.get("class", None)

        # we would like uri to be the highest dpi version so that
        # latex etc will use that.  But for now, lets just make
        # imagenm

        image_node["uri"] = imagenm
        image_node["srcset"] = self.options.get("srcset", None)

        return [image_node]


def _parse_srcset(st):
    """Parse st."""
    entries = st.split(",")
    srcset = {}
    for entry in entries:
        spl = entry.strip().split(" ")
        if len(spl) == 1:
            srcset[0] = spl[0]
        elif len(spl) == 2:
            mult = spl[1][:-1]
            srcset[float(mult)] = spl[0]
        else:
            raise ExtensionError('srcset argument "{entry}" is invalid.')
    return srcset


def visit_imgsg_html(self, node):
    """Handle HTML image tag depending on 'srcset' configuration.

    If 'srcset' is not `None`, copy images, generate image html tag with 'srcset'
    and add to HTML `body`. If 'srcset' is `None` run `visit_image` on `node`.
    """
    if node["srcset"] is None:
        self.visit_image(node)
        return

    imagedir, srcset = _copy_images(self, node)

    # /doc/examples/subd/plot_1.rst
    docsource = self.document["source"]
    # /doc/
    # make sure to add the trailing slash:
    srctop = os.path.join(self.builder.srcdir, "")
    # examples/subd/plot_1.rst
    relsource = os.path.relpath(docsource, srctop)
    # /doc/build/html
    desttop = os.path.join(self.builder.outdir, "")
    # /doc/build/html/examples/subd
    dest = os.path.join(desttop, relsource)

    # ../../_images/ for dirhtml and ../_images/ for html
    imagerel = os.path.relpath(imagedir, os.path.dirname(dest))
    if self.builder.name == "dirhtml":
        imagerel = os.path.join("..", imagerel, "")
    else:  # html
        imagerel = os.path.join(imagerel, "")

    if "\\" in imagerel:
        imagerel = imagerel.replace("\\", "/")
    # make srcset str.  Need to change all the prefixes!
    srcsetst = ""
    for mult in srcset:
        nm = os.path.basename(srcset[mult][1:])
        # ../../_images/plot_1_2_0x.png
        relpath = imagerel + nm
        srcsetst += f"{relpath}"
        if mult == 0:
            srcsetst += ", "
        else:
            srcsetst += f" {mult:1.2f}x, "
    # trim trailing comma and space...
    srcsetst = srcsetst[:-2]

    # make uri also be relative...
    nm = os.path.basename(node["uri"][1:])
    uri = imagerel + nm

    alt = node["alt"]
    if node["class"] is not None:
        classst = node["class"][0]
        classst = f'class = "{classst}"'
    else:
        classst = ""

    html_block = f'<img src="{uri}" srcset="{srcsetst}" alt="{alt}"' + f" {classst}/>"
    self.body.append(html_block)


def visit_imgsg_latex(self, node):
    """Copy images, set node[uri] to highest resolution image and call `visit_image`."""
    if node["srcset"] is not None:
        imagedir, srcset = _copy_images(self, node)
        maxmult = -1
        # choose the highest res version for latex:
        for key in srcset.keys():
            maxmult = max(maxmult, key)
        node["uri"] = str(PurePosixPath(srcset[maxmult]).name)

    self.visit_image(node)


def _copy_images(self, node):
    srcset = _parse_srcset(node["srcset"])

    # where the sources are.  i.e. myproj/source
    srctop = self.builder.srcdir

    # copy image from source to imagedir.  This is
    # *probably* supposed to be done by a builder but...
    # ie myproj/build/html/_images
    imagedir = os.path.join(self.builder.imagedir, "")
    imagedir = PurePosixPath(self.builder.outdir, imagedir)

    os.makedirs(imagedir, exist_ok=True)

    # copy all the sources to the imagedir:
    for mult in srcset:
        abspath = PurePosixPath(srctop, srcset[mult][1:])
        shutil.copyfile(abspath, imagedir / abspath.name)

    return imagedir, srcset


def depart_imgsg_html(self, node):
    """HTML depart node visitor function."""
    pass


def depart_imgsg_latex(self, node):
    """LaTeX depart node visitor function."""
    self.depart_image(node)


def imagesg_addnode(app):
    """Add `imgsgnode` to Sphinx app with visitor functions for HTML and LaTeX."""
    app.add_node(
        imgsgnode,
        html=(visit_imgsg_html, depart_imgsg_html),
        latex=(visit_imgsg_latex, depart_imgsg_latex),
    )



"""BlockParser divides non `.py` source files into blocks of code and markup text."""

import ast
import codecs
import re
from pathlib import Path
from textwrap import dedent

import pygments.lexers
import pygments.token
from sphinx.errors import ExtensionError
from sphinx.util.logging import getLogger

from .py_source_parser import FLAG_BODY, Block

logger = getLogger("sphinx-gallery")

# Don't just use "x in pygments.token.Comment" because it also includes preprocessor
# statements
COMMENT_TYPES = (
    pygments.token.Comment.Single,
    pygments.token.Comment,
    pygments.token.Comment.Multiline,
)


class BlockParser:
    """
    A parser that breaks a source file into blocks of code and markup text.

    Determines the source language and identifies comment blocks using pygments.

    Parameters
    ----------
    source_file : str
        A file name that has a suffix compatible with files that are subsequently parsed
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery.
    """

    def __init__(self, source_file, gallery_conf):
        source_file = Path(source_file)
        if name := gallery_conf["filetype_parsers"].get(source_file.suffix):
            self.lexer = pygments.lexers.find_lexer_class_by_name(name)()
        else:
            self.lexer = pygments.lexers.find_lexer_class_for_filename(source_file)()
        self.language = self.lexer.name

        # determine valid comment syntaxes. For each possible syntax, the tuple contains
        # - A test comment
        # - The comment start character
        # - The comment end character (for multiline comments) or None
        # - A regex for the start of a block based on repeated characters, or None
        comment_tests = [
            ("#= comment =#", "#=", "=#", None),  # Julia multiline
            ("# comment", "#", None, "#{20,}"),  # Julia, Ruby, Bash, Perl, etc.
            ("// comment", "//", None, "/{20,}"),  # C++, C#, Java, Rust, etc.
            ("/* comment */", r"/\*", r"\*/", r"/\*{20,}/"),  # C/C++ etc. multiline
            ("% comment", "%", None, "%{20,}"),  # Matlab
            ("! comment", "!", None, "!{20,}"),  # Fortran
            ("c     comment", r"^c(?:$|     )", None, None),  # Fortran 77
        ]

        self.allowed_comments = []
        allowed_special = []
        self.multiline_end = re.compile(chr(0))  # unmatchable regex
        for test, start, end, special in comment_tests:
            if next(self.lexer.get_tokens(test))[0] in COMMENT_TYPES:
                self.allowed_comments.append(start)
                if end:
                    self.multiline_end = re.compile(rf"(.*?)\s*{end}")
                if special:
                    allowed_special.append(special)

        if self.language == "Matlab":
            # Matlab treats code sections starting with "%%" in a similar manner to
            # tools that recognize "# %%" in Python code, so we accept that style as
            # an alternative.
            allowed_special.append(r"%%(?:$|\s)")

        if r"/\*" in self.allowed_comments:
            # Remove decorative asterisks and comment starts from C-style multiline
            # comments
            self.multiline_cleanup = re.compile(r"\s*/?\*\s*")
        else:
            self.multiline_cleanup = re.compile(r"\s*")

        comment_start = "|".join(self.allowed_comments)
        allowed_special = "|".join(allowed_special)
        if allowed_special:
            self.start_special = re.compile(
                f"(?:(?:{comment_start}) ?%% ?|{allowed_special})(.*)"
            )
        else:
            self.start_special = re.compile(f"(?:{comment_start}) ?%% ?(.*)")
        self.continue_text = re.compile(f"(?:{comment_start}) ?(.*)")

        # The pattern for in-file config comments is designed to not greedily match
        # newlines at the start and end, except for one newline at the end. This
        # ensures that the matched pattern can be removed from the code without
        # changing the block structure; i.e. empty newlines are preserved, e.g. in
        #
        #     a = 1
        #
        #     # sphinx_gallery_thumbnail_number = 2
        #
        #     b = 2
        flag_start = rf"^[\ \t]*(?:{comment_start})\s*"

        self.infile_config_pattern = re.compile(flag_start + FLAG_BODY, re.MULTILINE)
        self.start_ignore_flag = flag_start + "sphinx_gallery_start_ignore"
        self.end_ignore_flag = flag_start + "sphinx_gallery_end_ignore"
        self.ignore_block_pattern = re.compile(
            rf"{self.start_ignore_flag}(?:[\s\S]*?){self.end_ignore_flag}\n?",
            re.MULTILINE,
        )

    def split_code_and_text_blocks(self, source_file, return_node=False):
        """Return list with source file separated into code and text blocks.

        Parameters
        ----------
        source_file : str
            Path to the source file.
        return_node : bool
            Ignored; returning an ast node is not supported

        Returns
        -------
        file_conf : dict
            File-specific settings given in source file comments as:
            ``# sphinx_gallery_<name> = <value>``
        blocks : list
            (label, content, line_number)
            List where each element is a tuple with the label ('text' or 'code'),
            the corresponding content string of block and the leading line number.
        node : None
            Returning an ast node is not supported.
        """
        with codecs.open(source_file, "r", "utf-8") as fid:
            content = fid.read()
        # change from Windows format to UNIX for uniformity
        content = content.replace("\r\n", "\n")
        return self._split_content(content)

    def _get_content_lines(self, content):
        """
        Combine individual tokens into lines.

        Use the first non-whitespace token (if any) as the characteristic token type for
        the line.
        """
        current_line = []
        line_token = pygments.token.Whitespace
        for token, text in self.lexer.get_tokens(content):
            if line_token == pygments.token.Whitespace:
                line_token = token

            if "\n" in text:
                text_lines = text.split("\n")
                # first item belongs to the previous line
                current_line.append(text_lines.pop(0))
                yield line_token, "".join(current_line)
                # last item belongs to the line after this token
                current_line = [text_lines.pop()]
                # Anything left is a block of lines to add directly
                for ln in text_lines:
                    if not ln.strip():
                        line_token = pygments.token.Whitespace
                    yield line_token, ln
                if not current_line[0].strip():
                    line_token = pygments.token.Whitespace
            else:
                current_line.append(text)

    def _get_blocks(self, content):
        """
        Generate a sequence of "blocks" from the lines in ``content``.

        Each block is a tuple of (label, content, line_number), matching the format
        ultimately returned by `split_code_and_text_blocks`.
        """
        start_text = self.continue_text  # No special delimiter needed for first block
        needs_multiline_cleanup = False

        def cleanup_multiline(lines):
            nonlocal needs_multiline_cleanup
            first_line = 1 if start_text == self.continue_text else 0
            longest = max(len(line) for line in lines)
            matched = False
            for i, line in enumerate(lines[first_line:]):
                if m := self.multiline_cleanup.match(line):
                    matched = True
                    if (n := len(m.group(0))) < len(line):
                        longest = min(longest, n)

            if matched and longest:
                for i, line in enumerate(lines[first_line:], start=first_line):
                    lines[i] = lines[i][longest:]
            needs_multiline_cleanup = False
            return lines

        def finalize_block(mode, block):
            nonlocal start_text
            if mode == "text":
                if needs_multiline_cleanup:
                    cleanup_multiline(block)

                # subsequent blocks need to have the special delimiter
                start_text = self.start_special

                # Remove leading blank lines, and end in a single newline
                first = 0
                for i, line in enumerate(block):
                    first = i
                    if line.strip():
                        break
                last = None
                for i, line in enumerate(reversed(block)):
                    last = -i or None
                    if line.strip():
                        break
                block.append("")
                text = dedent("\n".join(block[first:last]))
            else:
                text = "\n".join(block)
            return Block(mode, text, n - len(block))

        block = []
        mode = None
        for n, (token, text) in enumerate(self._get_content_lines(content)):
            if mode == "text" and token in pygments.token.Whitespace:
                # Blank line ends current text block
                if block:
                    yield finalize_block(mode, block)
                mode, block = None, []
            elif (
                mode != "text"
                and token in COMMENT_TYPES
                and (m := start_text.search(text))
            ):
                # start of a text block; end the current block
                if block:
                    yield finalize_block(mode, block)
                mode, block = "text", []
                if (trailing_text := m.group(1)) is not None:
                    if start_text == self.continue_text:
                        if (delimited := self.start_special.search(text)) and (
                            heading := delimited.group(1).strip()
                        ):
                            # Treat text on the same line as the delimiter as a title
                            block.extend((heading, "=" * len(heading), ""))
                        else:
                            # Keep any text on the first line of the title block as is
                            block.append(trailing_text)
                    elif heading := trailing_text.strip():
                        # Treat text on the same line as the delimiter as a heading
                        block.extend((heading, "-" * len(heading), ""))
            elif mode == "text" and token in COMMENT_TYPES:
                # Continuation of a text block
                if m := self.start_special.search(text):
                    # Starting a new text block now is just a continuation of the
                    # existing one, possibly including a new heading.
                    if heading := m.group(1).strip():
                        block.extend((heading, "-" * len(heading), ""))
                    pass
                elif token == pygments.token.Comment.Multiline:
                    if m := self.multiline_end.search(text):
                        block.append(m.group(1))
                        needs_multiline_cleanup = True
                    else:
                        block.append(text)
                else:
                    block.append(self.continue_text.search(text).group(1))
            elif mode != "code":
                # start of a code block
                if block:
                    yield finalize_block(mode, block)
                mode, block = "code", [text]
            else:
                # continuation of a code block
                block.append(text)

        # end of input ends final block
        if block:
            yield finalize_block(mode, block)

    def _split_content(self, content):
        """
        Split the input content into blocks.

        Return a tuple of (file_conf, blocks, None) that corresponds to the return
        values of ``split_code_and_text_blocks``.
        """
        file_conf = self.extract_file_config(content)
        blocks = list(self._get_blocks(content))

        # For examples that start with a code block before the file docstring due to
        # language conventions or requirements, swap these blocks so the title block
        # comes first and merge consecutive code blocks if needed.
        if len(blocks) >= 2 and blocks[0].type == "code" and blocks[1].type == "text":
            blocks[0], blocks[1] = blocks[1], blocks[0]
            if len(blocks) >= 3 and blocks[2].type == "code":
                blocks[1] = Block(
                    "code",
                    f"{blocks[1].content}\n{blocks[2].content}",
                    blocks[1].lineno,
                )
                blocks.pop(2)

        return file_conf, blocks, None

    def extract_file_config(self, content):
        """Pull out the file-specific config specified in the docstring."""
        file_conf = {}
        for match in re.finditer(self.infile_config_pattern, content):
            name = match.group(1)
            value = match.group(3)
            if value is None:  # a flag rather than a config setting
                continue
            try:
                value = ast.literal_eval(value)
            except (SyntaxError, ValueError):
                logger.warning(
                    "Sphinx-gallery option %s was passed invalid value %s", name, value
                )
            else:
                file_conf[name] = value
        return file_conf

    def remove_ignore_blocks(self, code_block):
        """
        Return the content of *code_block* with ignored areas removed.

        An ignore block starts with ``?? sphinx_gallery_start_ignore`` and ends with
        ``?? sphinx_gallery_end_ignore`` where ``??`` is the active language's line
        comment marker. These lines and anything in between them will be removed, but
        surrounding empty lines are preserved.

        Parameters
        ----------
        code_block : str
            A code segment.
        """
        num_start_flags = len(re.findall(self.start_ignore_flag, code_block))
        num_end_flags = len(re.findall(self.end_ignore_flag, code_block))

        if num_start_flags != num_end_flags:
            raise ExtensionError(
                'All "sphinx_gallery_start_ignore" flags must have a matching '
                '"sphinx_gallery_end_ignore" flag!'
            )
        return re.subn(self.ignore_block_pattern, "", code_block)[0]

    def remove_config_comments(self, code_block):
        """
        Return the content of *code_block* with in-file config comments removed.

        Comment lines with the pattern ``sphinx_gallery_[option] = [val]`` after the
        line comment character are removed, but surrounding empty lines are preserved.

        Parameters
        ----------
        code_block : str
            A code segment.
        """
        parsed_code, _ = re.subn(self.infile_config_pattern, "", code_block)
        return parsed_code



"""
Logging test
============
"""

# %%

import logging
import sys


def _create_logger(name=None):
    logger = logging.getLogger(name=name)
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    logger.addHandler(sh)
    return logger


# %%
logger = _create_logger("first_logger")
logger.info("is in the same cell")

# %%
logger.info("is not in the same cell")



r"""
Seaborn example
===============

This example demonstrates a Seaborn plot. Figures produced Matplotlib **and**
by any package that is based on Matplotlib (e.g., Seaborn), will be
captured by default. See :ref:`image_scrapers` for details.
"""

# Author: Michael Waskom & Lucy Liu
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# Enforce the use of default set style

# Create a noisy periodic dataset
y_array = np.array([])
x_array = np.array([])
rs = np.random.RandomState(8)
for _ in range(15):
    x = np.linspace(0, 30 / 2, 30)
    y = np.sin(x) + rs.normal(0, 1.5) + rs.normal(0, 0.3, 30)
    y_array = np.append(y_array, y)
    x_array = np.append(x_array, x)

# Plot the average over replicates with confidence interval
sns.lineplot(y=y_array, x=x_array)
# to avoid text output
plt.show()



"""
Providing a figure for the thumbnail image
==========================================

This example demonstrates how to provide a figure that is displayed as the
thumbnail. This is done by specifying the keyword-value pair
``sphinx_gallery_thumbnail_path = 'fig path'`` as a comment somewhere below the
docstring in the example file. In this example, we specify that we wish the
figure ``demo.png`` in the folder ``_static`` to be used for the thumbnail.
"""

import matplotlib.pyplot as plt
import numpy as np

# sphinx_gallery_thumbnail_path = '_static/demo.png'

# %%

x = np.linspace(0, 4 * np.pi, 301)
y1 = np.sin(x)
y2 = np.cos(x)

# %%
# Plot 1
# ------

plt.figure()
plt.plot(x, y1, label="sin")
plt.plot(x, y2, label="cos")
plt.legend()
plt.show()

# %%
# Plot 2
# ------

plt.figure()
plt.plot(x, y1, label="sin")
plt.plot(x, y2, label="cos")
plt.legend()
plt.xscale("log")
plt.yscale("log")
plt.show()



"""Test utility functions."""

from pathlib import Path

from sphinx_gallery.utils import _combine_backreferences, _read_json, _write_json


def test_combine_backreferences():
    """Check `_combine_backreferences` works as expected."""
    backrefs_a = {
        "a": [1, 2],
        "b": [11, 12],
    }
    assert _combine_backreferences({}, backrefs_a) == backrefs_a

    backrefs_b = {
        "a": [3, 4],
        "c": [21, 22],
    }
    assert _combine_backreferences(backrefs_a, backrefs_b) == {
        "a": [1, 2, 3, 4],
        "b": [11, 12],
        "c": [21, 22],
    }


def test_read_write_json(tmpdir):
    """Check `_read_json` and `_write_json` work as expected."""
    path = Path(tmpdir, "test")
    data = {
        "object1": ("path/file.py", "first intro", "first title"),
        "object2": ("path2/file2.py", "second intro", "second title"),
    }
    _write_json(path, data, "test_dict")
    # Writing converts tuples to lists
    assert _read_json(Path(path).with_name(path.stem + "test_dict.json")) == {
        key: list(value) for key, value in data.items()
    }



"""Utility functions for doc building."""

import os.path as op

from sphinx_gallery.scrapers import matplotlib_scraper


def notebook_modification_function(notebook_content, notebook_filename):
    """Implement JupyterLite-specific modifications of notebooks."""
    source = f"JupyterLite-specific change for {notebook_filename}"
    markdown_cell = {"cell_type": "markdown", "metadata": {}, "source": source}
    notebook_content["cells"] = [markdown_cell] + notebook_content["cells"]


class MatplotlibFormatScraper:
    """Calls Matplotlib scraper, passing required `format` kwarg for testing."""

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, block, block_vars, gallery_conf):
        """Call Matplotlib scraper with required `format` kwarg for testing."""
        kwargs = dict()
        if (
            op.basename(block_vars["target_file"]) == "plot_svg.py"
            and gallery_conf["builder_name"] != "latex"
        ):
            kwargs["format"] = "svg"
        elif (
            op.basename(block_vars["target_file"]) == "plot_webp.py"
            and gallery_conf["builder_name"] != "latex"
        ):
            kwargs["format"] = "webp"
        return matplotlib_scraper(block, block_vars, gallery_conf, **kwargs)


class ResetArgv:
    """Provide `reset_argv` callable returning required `sys.argv` for test."""

    def __repr__(self):
        return "ResetArgv"

    def __call__(self, sphinx_gallery_conf, script_vars):
        """Return 'plot' arg if 'plot_command_line_args' example, for testing."""
        if "plot_command_line_args.py" in script_vars["src_file"]:
            return ["plot"]
        else:
            return []


def _raise(*args, **kwargs):
    import matplotlib.pyplot as plt

    plt.close("all")
    raise ValueError(
        "zero-size array to reduction operation minimum which has no identity"
    )


class MockScrapeProblem:
    """Used in 'reset_modules' to mock error during scraping."""

    def __init__(self):
        from matplotlib.colors import colorConverter

        self._orig = colorConverter.to_rgba

    def __repr__(self):
        return "MockScrapeProblem"

    def __call__(self, gallery_conf, fname):
        """Raise error for 'scraper_broken' example."""
        from matplotlib.colors import colorConverter

        if "scraper_broken" in fname:
            colorConverter.to_rgba = _raise
        else:
            colorConverter.to_rgba = self._orig


class MockSort:
    """Fake sort used to test that mini-gallery sort is correct."""

    def __repr__(self):
        return "MockSort"

    def __call__(self, f):
        """Sort plot_sub* for one test case."""
        if "subdir2" in f:
            return 0
        if "subdir1" in f:
            return 1
        return f


mock_scrape_problem = MockScrapeProblem()
matplotlib_format_scraper = MatplotlibFormatScraper()
reset_argv = ResetArgv()
mock_sort = MockSort()


def noop_key(x):
    """Sortkey that passes x."""
    return x



"""
Constrained Text output frame
=============================

This example captures the standard output and includes it in the
example. If output is too long it becomes automatically
framed into a text area.
"""

# Code source: Óscar Nájera
# License: BSD 3 clause

print("This is a long test Output\n" * 50)

####################################
# One line out

print("one line out")
